/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: Statistics.hpp 2167 2017-03-01 12:10:04Z stephane $
 * See license.txt.
 */

#pragma once

#include "common.hpp"
#include <fstream>


namespace EPS
{
	/// Track all of the transfer and bandwidth statistics.
	struct Statistics final
	{
		public:

			uint64_t server_packets_forwarded;	///< packets forwarded by the server to a client
			uint64_t server_bytes_forwarded;	///< bytes forwarded by the server to a client
			uint64_t server_packets_received;	///< packets received from the server
			uint64_t server_bytes_received;		///< bytes received from the server
			uint64_t server_packets_sent;		///< packets sent to the server
			uint64_t server_bytes_sent;			///< bytes sent to the server

			uint64_t packets_dropped;			///< packets dropped
			uint64_t bytes_dropped;				///< bytes dropped

			uint64_t direct_packets_received;	///< direct client-to-client packets received
			uint64_t direct_bytes_received;		///< direct client-to-client bytes received
			uint64_t direct_packets_sent;		///< direct client-to-client packets sent
			uint64_t direct_bytes_sent;			///< direct client-to-client bytes sent

			uint64_t tap_packets_received;		///< packets read from the TAP interface
			uint64_t tap_bytes_received;			///< bytes read from the TAP interface
			uint64_t tap_packets_sent;			///< packets written to the TAP interface
			uint64_t tap_bytes_sent;				///< bytes written to the TAP interface

			uint64_t connections_established;
			uint64_t connections_dropped;

			/// The BPS strings are calculated from other variables, and represent the latest results, not totals. @{
			char server_bps_in	[EPS::bps_string_size];
			char server_bps_out	[EPS::bps_string_size];
			char server_bps_fwd	[EPS::bps_string_size];
			char direct_bps_in	[EPS::bps_string_size];
			char direct_bps_out	[EPS::bps_string_size];
			char tap_bps_in		[EPS::bps_string_size];
			char tap_bps_out	[EPS::bps_string_size];
			/// @}

			// NOTE #1:  if you add new fields, please remember to update operator+=()
			// NOTE #2:  all fields must be POD because of the memset() in clear() -- do not use objects like std::string!

			/// Constructor.
			Statistics(void);

			/// Erase all stats.
			void clear(void);

			/// Combine all statistics from rhs into lhs.
			Statistics &operator+=(const Statistics &rhs);

			/// Convert the statistics to a text string for logging.
			std::string to_string(void) const;

			/// Determine if the statistics are all zero.
			bool empty(void) const;

			/// Export the statistics to a scorecard.
			void export_scorecard(std::ofstream &output, const std::string &prefix="") const;
	};


	/// Manage the @ref EPS::Statistics structure.  Differentiate between total statistics, and those gathered in the last minute.
	class Stats final
	{
		public:

			/// Constructor.
			Stats(void);

			/// Convert the statistics to a text string for logging.
			std::string to_string(void) const;

			/// Determine if the statistics are all zero.
			bool empty(void) const;

			/// Roll the minute stats into the total stats.
			void roll(void);

			EPS::Statistics minute_stats;
			EPS::Statistics total_stats;

			bool		log_stats;
			bool		scorecard_enabled;
			std::string	scorecard_filename;
	};
};
