/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: logging.cpp 2167 2017-03-01 12:10:04Z stephane $
 * See license.txt.
 */

#include "logging.hpp"
#include "common.hpp"
#include <boost/regex.hpp>
#include <boost/log/core.hpp>
#include <boost/log/trivial.hpp>
#include <boost/log/expressions.hpp>
#include <boost/log/utility/setup.hpp>
#include <boost/log/support/date_time.hpp>
#include <iomanip>
#ifdef EPS_WINDOWS
	#include <DbgHelp.h>
	#pragma comment( lib, "dbghelp.lib" )
#else
	#include <execinfo.h>
#endif


void EPS::initialize_logging(const EPS::EMode mode)
{
	auto core_logger = boost::log::core::get();

	core_logger->set_logging_enabled(true);
	core_logger->add_global_attribute("Scope"		, boost::log::attributes::named_scope()			);
	core_logger->add_global_attribute("TimeStamp"	, boost::log::attributes::utc_clock()			);
	core_logger->add_global_attribute("ProcessID"	, boost::log::attributes::current_process_id()	);
	core_logger->add_global_attribute("ThreadID"	, boost::log::attributes::current_thread_id()	);

	auto console_logger = boost::log::add_console_log( std::clog );
	console_logger->set_filter( boost::log::trivial::severity >= boost::log::trivial::info );

	LOG_INFO << "logging initialized for EPS " << EPS::mode_to_string(mode) << " v" << EPS_VERSION;

	return;
}


void EPS::initialize_file_logging(const std::string &filename)
{
	auto file_logger = boost::log::add_file_log(
			boost::log::keywords::file_name				= filename,
			boost::log::keywords::open_mode				= std::ios::app,
			boost::log::keywords::auto_flush			= true,
#if 0	//	Rotation is now handled by logrotate (since I cannot seem to get this to work)
			boost::log::keywords::rotation_size			= 10 * 1024 * 1024,	// file size at which the log is rotated
			boost::log::keywords::max_size				= 30 * 1024 * 1024,	// maximum size of all rolled files
			boost::log::keywords::time_based_rotation	= boost::log::sinks::file::rotation_at_time_point(0, 0, 0),	// midnight
#endif
			boost::log::keywords::format				=
			(
				boost::log::expressions::stream
						<< boost::log::expressions::format_date_time<boost::posix_time::ptime>("TimeStamp", "%Y-%m-%d %H:%M:%S.%f")
						<< " ["						<< boost::log::expressions::attr<boost::log::attributes::current_process_id::value_type>("ProcessID")
						<< ":"						<< boost::log::expressions::attr<boost::log::attributes::current_thread_id::value_type>("ThreadID")
						<< "]" << std::setw(8)		<< boost::log::expressions::attr<boost::log::trivial::severity_level>("Severity")
						<< " "						<< boost::log::expressions::format_named_scope(
															"Scope",
															boost::log::keywords::format	= "%n (%f:%l)",
															boost::log::keywords::iteration	= boost::log::expressions::forward,
															boost::log::keywords::depth		= 3)
						<< " - "					<< boost::log::expressions::smessage
			) );

	try
	{
		file_logger->set_filter( boost::log::trivial::severity >= boost::log::trivial::trace );

		// insert a blank line into the log file to make it visually obvious where there was a restart
		LOG_TRACE	<< "-------------------------------" << std::endl;
		LOG_INFO	<< "file logging enabled to " << filename << " for EPS v" << EPS_VERSION;
	}
	catch (...)
	{
		boost::log::core::get()->remove_sink(file_logger);
		LOG_ERROR << "ERROR!  Failed to open or use \"" << filename << "\" for log output -- file logging will be disabled";
	}

	return;
}


void EPS::reset_logging(void)
{
	auto core_logger = boost::log::core::get();
	core_logger->set_logging_enabled(false);

	return;
}


void EPS::log_backtrace(void)
{
	#ifdef EPS_LINUX

		/// @todo why does this not compile in clang/pc-bsd?

		const int size_of_backtrace = 30;
		void *buffer[size_of_backtrace];

		const int len	= backtrace(buffer, size_of_backtrace);
		char **str		= backtrace_symbols(buffer, len);

		LOG_DEBUG << "BT: captured " << len << " items from the call stack; logging from least to most recent:";

		/* see if we can clean up the backtrace lines by demangling the C++ name; for example:
		 *
		 *		src-client/eps-client(_ZN3EPS9NetBuffer3getENS0_5ETypeE+0x12c) [0x5a3278]
		 *		src-client/eps-client(_ZN3EPS6Cipher14decrypt_headerERNS_9NetBufferE+0x33) [0x57a63f]
		 *		src-client/eps-client(_ZN3EPS13log_backtraceEv+0x2a) [0x5aee14]
		 *
		 * ...will get converted to:
		 *
		 *		src-client/eps-client(EPS::NetBuffer::get(EPS::NetBuffer::EType)+0x12c) [0x5a3278]
		 *		src-client/eps-client(EPS::Cipher::decrypt_header(EPS::NetBuffer&)+0x33) [0x57a63f]
		 *		src-client/eps-client(EPS::log_backtrace()+0x2a) [0x5aee14]
		 */
		boost::regex rx(
				"^"					// start of string
				"(.+\\()"			// 1st group up to and including the open parentheses
				"([^\\+\\)]+)"		// 2nd group is the mangled function name
				"(\\+0x[^\\)]+)"	// 3rd group is the offset into the function
				"(.+)$"				// 4th group is the rest of the line
				);

		for (int idx = len - 1; idx >= 0; idx --)
		{
			std::string text = str[idx];

			// must match the entire line if we want to try and demangle
			boost::smatch what;
			if (boost::regex_match(text, what, rx))
			{
				text = what[1] + " " + EPS::cpp_demangle(what[2]) + " " + what[3] + what[4];
			}

			LOG_DEBUG << "BT #" << idx << ": " << text;
		}

		free(str);

	#elif defined EPS_WINDOWS

		HANDLE process = GetCurrentProcess();
		static bool symbolsHaveBeenLoaded = false;
		if (symbolsHaveBeenLoaded == false)
		{
			// This can result in a search of many directories on disk.
			// Only do it the first time an exception is thrown, and then don't call SymCleanup().
			SymSetOptions(
				SYMOPT_CASE_INSENSITIVE			|   // case-insensitive searches
				SYMOPT_FAIL_CRITICAL_ERRORS		|   // do not display dialog boxes
				SYMOPT_INCLUDE_32BIT_MODULES	|   // include 32-bit modules
				SYMOPT_NO_PROMPTS				|   // prevent prompting for validation
				SYMOPT_UNDNAME					);  // use undecorated form

			if (SymInitialize(process, NULL, TRUE) == TRUE)
			{
				symbolsHaveBeenLoaded = true;
			}
		}

		const int size_of_backtrace = 30;
		void *stack[size_of_backtrace];

		const USHORT numberOfCapturedFrames = CaptureStackBackTrace(0, size_of_backtrace, stack, nullptr);

		LOG_DEBUG << "BT: captured " << numberOfCapturedFrames << " items from the call stack; logging from least to most recent:";

		char buffer[sizeof(SYMBOL_INFO) + MAX_SYM_NAME * sizeof(TCHAR)];
		SYMBOL_INFO *symbol = (SYMBOL_INFO*)buffer;
		symbol->MaxNameLen = MAX_SYM_NAME;
		symbol->SizeOfStruct = sizeof(SYMBOL_INFO);

		for (size_t idx = 0; idx < numberOfCapturedFrames; idx++)
		{
			std::string name = "unknown";
			if (SymFromAddr(process, DWORD64(stack[idx]), 0, symbol) == TRUE)
			{
				name = std::string(symbol->Name, symbol->NameLen) + "()";
			}
			LOG_DEBUG << "BT #" << idx << ": " << name;
		}

		/* We're keeping symbols loaded once done in case another exception
		 * is thrown or someone else tries to use the symbols.
		 * See "symbolsHaveBeenLoaded" above.
		SymCleanup();
		 */
		CloseHandle(process);

	#endif

	return;
}
