/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: Tracker.cpp 2210 2017-04-27 09:06:37Z stephane $
 * See license.txt.
 */

#include "Tracker.hpp"
#include "logging.hpp"
#include "Hull.hpp"
#include <chrono>
#include <boost/regex.hpp>
#include <boost/uuid/uuid_io.hpp>
#include <boost/uuid/string_generator.hpp>
#include <sys/stat.h>


std::string EPS::TrackerData::as_debug(void) const
{
	EPS::Tracker &tracker = EPS::Hull::get().tracker;

	std::stringstream ss;
	ss	<< "uuid: "				<< this->uuid
		<< ", mode: "			<< EPS::mode_to_string(this->mode)
		<< ", eps address: "	<< this->address_eps.to_string()
		<< ", other address: "	<< ipset_to_string(this->other_addresses)
		<< ", possible mesh: "	<< ipset_to_string(this->mesh_addresses)
		<< ", connection: "		<< this->connection
		<< ", hostname: "		<< this->hostname
		<< ", os name: "		<< this->os_name
		<< ", use count: "		<< this->connection.use_count()
		<< ", active: "			<< (this->active						? "yes" : "no")
		<< ", is_active: "		<< (this->is_active()					? "yes" : "no")
		<< ", trk_map[uuid]: "	<< (tracker.exists(this->uuid		)	? "yes" : "no")
		<< ", eps_map[addr]: "	<< (tracker.exists(this->address_eps)	? "yes" : "no")
		;

	if (not this->msg.empty())
	{
		ss << ", message: " << this->msg;
	}

	return ss.str();
}


EPS::TrackerData &EPS::Tracker::add_inactive_entry(const UUID &uuid, const IPv4 &address_eps)
{
	if (uuid.is_nil())
	{
		THROW_INVALID_ARGUMENT(-1, "Cannot add a tracker entry with a null uuid.");
	}

	if (this->exists(uuid))
	{
		LOG_TRACE << "updating entry for " << uuid << " aka " << address_eps.to_string();
	}
	else
	{
		LOG_TRACE << "creating new entry for " << uuid << " aka " << address_eps.to_string();
	}

	// this will create a new entry if necessary, or update an existing entry
	EPS::TrackerData &data			= this->get(uuid);
	data.active						= false;
	data.uuid						= uuid;

	this->update_entry(uuid);

	if (not address_eps.is_unspecified())
	{
		data.address_eps = address_eps;
		this->eps_address_map[data.address_eps] = uuid;
	}

	return data;
}


EPS::TrackerData &EPS::Tracker::add_active_entry(const UUID &uuid, const IPv4 &address_eps, const EPS::TConnPtr &connection)
{
	LOG_TRACE << "adding an active entry for " << uuid << " at " << address_eps.to_string();

	this->add_inactive_entry(uuid, address_eps);

	if (connection == nullptr)
	{
		THROW_INVALID_ARGUMENT(-1, "Cannot activate a tracker entry with a null connection.");
	}
	if (connection->uuid != uuid && not connection->uuid.is_nil())
	{
		// something is wrong -- the connection's UUID is already set to something else
		THROW_RUNTIME_ERROR(-1, "UUID for connection to " + address_eps.to_string() + " has already been set to something else.");
	}

	connection->uuid = uuid;

	return this->activate_entry(connection);
}


EPS::TrackerData &EPS::Tracker::activate_entry(const EPS::TConnPtr &connection)
{
	if (connection.get() == nullptr)
	{
		THROW_INVALID_ARGUMENT(-1, "Cannot activate a tracker entry with a null connection.");
	}

	const UUID uuid = connection->uuid;
	if (uuid.is_nil())
	{
		THROW_INVALID_ARGUMENT(-1, "Cannot activate a tracker entry for " + connection->name + " with a null uuid.");
	}

	EPS::TrackerData &data	= this->get(uuid);

	LOG_TRACE << "activating entry: " << data.as_debug();

	data.active					= connection->socket.is_open();
	data.connection				= connection;
	data.disconnect_timestamp	= 0;
	data.other_addresses.insert( connection->endpoint.address().to_v4() );
	data.pending_mesh_connection.reset();

	if (not data.address_eps.is_unspecified())
	{
		this->eps_address_map[data.address_eps] = uuid;
	}

	LOG_TRACE << "activated entry: " << data.as_debug();

	return data;
}


EPS::TrackerData &EPS::Tracker::deactivate_entry(const EPS::TConnPtr &connection)
{
	// the connection *must* exist otherwise get() will throw

	EPS::TrackerData &data = this->get(connection);

	return this->deactivate_entry(data.uuid);
}


EPS::TrackerData &EPS::Tracker::deactivate_entry(const UUID &uuid)
{
	// We'll remember the UUID and addresses in case this connection comes back, but otherwise forget everything else.
	//
	// This method may be called on a UUID that the tracker hasn't heard about yet.  (For example, during the
	// "validating connection" phase prior to a connection having been inserted into Tracker.)

	EPS::TrackerData &data = this->get(uuid);

	std::string name;
	if (data.connection)
	{
		name = data.connection->name;
	}
	LOG_TRACE << name << "de-activate entry for " << uuid << " (usage=" << data.connection.use_count() << ")";

	this->eps_address_map.erase(data.address_eps);

	if (data.connection)
	{
		data.connection->uuid = boost::uuids::nil_uuid();
	}

	data.connection				.reset();
	data.pending_mesh_connection.reset();
	data.mesh_addresses			.clear();
	data.active					= false;

	if (data.disconnect_timestamp == 0)
	{
		data.disconnect_timestamp = std::time(nullptr);
	}

	return data;
}


EPS::TrackerData &EPS::Tracker::update_entry(const UUID &uuid)
{
	EPS::TrackerData &data = this->get(uuid);

	const auto now = std::chrono::system_clock::now();
	data.last_updated				= now;
	data.next_connection_attempt	= now;

	return data;
}


bool EPS::Tracker::exists(const UUID &uuid) const
{
	return (this->trk_map.count(uuid) == 1);
}


bool EPS::Tracker::exists(const IPv4 &eps_ipv4) const
{
	return (this->eps_address_map.count(eps_ipv4) == 1);
}


bool EPS::Tracker::exists(const EPS::TConnPtr &connection) const
{
	if (connection)
	{
		if (not connection->uuid.is_nil())
		{
			return this->exists(connection->uuid);
		}

		// if the uuid hasn't yet been set on the connection, look through all entries
		for (auto iter : this->trk_map)
		{
			const TrackerData &data = iter.second;
			if (data.connection					== connection ||
				data.pending_mesh_connection	== connection )
			{
				return true;
			}
		}
	}

	return false;
}


EPS::TrackerData &EPS::Tracker::get(const UUID &uuid)
{
	EPS::TrackerData &data = this->trk_map[uuid];
	if (data.uuid.is_nil())
	{
		// this must be a new entry
		data.uuid			= uuid;
		data.last_updated	= std::chrono::system_clock::now();
	}

	return data;
}


EPS::TrackerData &EPS::Tracker::get(const IPv4 &eps_ipv4)
{
	return this->get(this->eps_address_map[eps_ipv4]);
}


EPS::TrackerData &EPS::Tracker::get(const TConnPtr &connection)
{
	if (not connection)
	{
		THROW_INVALID_ARGUMENT(-1, "connection pointer cannot be null");
	}

	if (not connection->uuid.is_nil())
	{
		return this->get(connection->uuid);
	}

	// if the uuid hasn't yet been set on the connection, look through all entries
	for (auto &iter : this->trk_map)
	{
		TrackerData &data = iter.second;
		if (data.connection					== connection ||
			data.pending_mesh_connection	== connection )
		{
			return data;
		}
	}

	THROW_RUNTIME_ERROR(-1, "failed to find a valid connection");
}


EPS::TConnPtr EPS::Tracker::create_shared_pointer(EPS::TCPConnection *connection)
{
	if (connection == nullptr)
	{
		THROW_INVALID_ARGUMENT( -1, "Cannot create a shared pointer for \"nullptr\"" );
	}

	// first check to see if a shared pointer already exists...
	for ( auto &tconn : this->get_all_connections() )
	{
		if (tconn.get() == connection)
		{
			return tconn;
		}
	}

	// ...otherwise we create a new one
	EPS::TConnPtr ptr(connection);
	return ptr;
}


EPS::TConnPtr EPS::Tracker::get_shared_pointer(EPS::TCPConnection *connection)
{
	if (connection == nullptr)
	{
		THROW_INVALID_ARGUMENT( -1, "Cannot find a shared pointer for \"nullptr\"" );
	}

	for ( auto &tconn : this->get_all_connections() )
	{
		if ( tconn.get() == connection )
		{
			// we found the right shared pointer!
			return tconn;
		}
	}

	// when a connection to the server is first established, it is possible we have
	// a shared pointer that isn't yet in the tracker -- see if that is the case

	EPS::Hull &hull = EPS::Hull::get();
	for (auto &tconn : hull.server_conns)
	{
		if ( tconn.get() == connection )
		{
			return tconn;
		}
	}

	// Is "connection" even a valid pointer?  Why do we not have a shared pointer for it?
	THROW_INVALID_ARGUMENT(-1, "Cannot find a shared pointer for connection");
}


void EPS::Tracker::deactivate_old_entries(void)
{
	const std::chrono::system_clock::time_point deadline = std::chrono::system_clock::now() - std::chrono::minutes(3);

	// first pass:  find all entries older than "deadline"
	UUIDSet uuids_to_deactivate;
	for (const auto &iter : this->trk_map)
	{
		const EPS::TrackerData &data = iter.second;
		if (data.last_updated < deadline)
		{
			uuids_to_deactivate.insert(data.uuid);
		}
	}

	// second pass:  de-activate all entries identified in the first pass
	while (not uuids_to_deactivate.empty())
	{
		const UUID uuid = *uuids_to_deactivate.begin();
		uuids_to_deactivate.erase(uuid);
		LOG_TRACE << "de-activate " << uuid << " from tracker (entry too old)";
		this->deactivate_entry(uuid);
	}

	return;
}


EPS::UUID EPS::Tracker::uuid_for_destination(const IPv4 &eps_ipv4) const
{
	/// @returns nil UUID if the destination IPv4 address is unknown
	UUID uuid = boost::uuids::nil_uuid();

	if (eps_address_map.count(eps_ipv4) != 0)
	{
		/// @returns UUID corresponding to the destination %EPS IP address
		uuid = eps_address_map.at(eps_ipv4);
	}

	return uuid;
}


EPS::TCPConns EPS::Tracker::get_all_connections(void) const
{
	const EPS::TConnPtr ptr;
	return this->get_all_connections_other_than(ptr);
}


EPS::TCPConns EPS::Tracker::get_all_connections_other_than(const TConnPtr &connection) const
{
	EPS::TCPConns conns;

	for (const auto iter : this->trk_map)
	{
		const TrackerData &data = iter.second;

		if (data.connection != connection &&
			data.connection.get() != nullptr)
		{
			conns.insert(data.connection);
		}
		else
		{
			if (data.pending_mesh_connection != connection &&
				data.pending_mesh_connection.get() != nullptr)
			{
				conns.insert(data.pending_mesh_connection);
			}
		}
	}

	return conns;
}


EPS::UUIDSet EPS::Tracker::get_all_uuids(void) const
{
	EPS::UUIDSet uuids;

	for (const auto iter : this->trk_map)
	{
		const TrackerData &data = iter.second;

		uuids.insert(data.uuid);
	}

	return uuids;
}


EPS::UUIDSet EPS::Tracker::get_all_active_uuids(void) const
{
	EPS::UUIDSet uuids;

	for (const auto iter : this->trk_map)
	{
		const TrackerData &data = iter.second;

		if (data.is_active())
		{
			uuids.insert(data.uuid);
		}
	}

	return uuids;
}


void EPS::Tracker::export_to_cache(const IPv4 &network) const
{
	if (this->trk_map.empty())
	{
		// nothing to save
		return;
	}

	#ifndef EPS_WINDOWS
		const std::string directory	= "/var/cache/eps";
		const std::string filename	= directory + "/" + network.to_string() + ".cache";

		mkdir(directory.c_str(), S_IRWXU);

		std::ofstream cache(filename, std::ofstream::out | std::ofstream::trunc);
		if (cache.bad() || cache.fail() || not cache.is_open())
		{
			LOG_ERROR << "failed to export server's cache to " << filename;
			return;
		}

		cache << "# EPS Conduits v" << EPS_VERSION << std::endl;

		for (const auto &iter : this->trk_map)
		{
			const EPS::TrackerData &data = iter.second;

			cache	<< data.uuid					<< " "
					<< data.address_eps.to_string()	<< " "
													<< std::endl;
		}

		LOG_TRACE << "saved " << this->trk_map.size() << " entries to " << filename;
	#else
		/// @todo where to save this in Windows?
	#endif

	return;
}


void EPS::Tracker::import_from_cache(const IPv4 &network)
{
	const std::string directory	= "/var/cache/eps";
	const std::string filename	= directory + "/" + network.to_string() + ".cache";

	std::ifstream cache(filename, std::ofstream::in);
	if (cache.bad() || cache.fail() || not cache.is_open())
	{
		LOG_TRACE << "cannot import tracker data from " << filename;
		return;
	}

	const boost::regex rx(
		"^"					// start of line
		"([-A-Za-z0-9]{36})"// uuid
		" "					// space
		"([0-9\\.]{7,15})"	// esp ip address
		".*"				// ignore any other field
		"$"					// end of line
	);

	auto gen = boost::uuids::string_generator();

	std::string line;
	while (std::getline(cache, line))
	{
		boost::smatch what;
		if (not boost::regex_match(line, what, rx))
		{
			// line doesn't match -- ignore
			continue;
		}

		const std::string uuid_str = what[1].str();
		const std::string addr_str = what[2].str();

		const UUID uuid = gen(uuid_str);
		const IPv4 addr = IPv4::from_string(addr_str);

		this->add_inactive_entry(uuid, addr);
	}

	LOG_INFO << "imported " << this->trk_map.size() << " entries from cache for network " << network;

	return;
}
