/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: Tracker.hpp 2186 2017-04-01 17:47:29Z stephane $
 * See license.txt.
 */

#pragma once

#include "common.hpp"
#include "TCPConnection.hpp"
#include <boost/uuid/uuid.hpp>
#include <boost/uuid/nil_generator.hpp>
#include <chrono>
#include <map>


namespace EPS
{
	/// Used to track everything we know about a particular %EPS client.
	struct TrackerData final
	{
		// members

		bool									active;						///< active connection (can be used to send/receive packets)
		EPS::EMode								mode;						///< type of peer (client or server?)
		UUID									uuid;						///< uuid for this %EPS conduit
		IPv4									address_eps;				///< %EPS address last assigned to this client
		IPv4Set									other_addresses;			///< all other (non-%EPS) addresses where this device can be reached
		IPv4Set									mesh_addresses;				///< set of addresses to try when connecting directly
		EPS::TConnPtr							connection;					///< shared pointer to the @b direct peer tcp connection (can be empty)
		EPS::TConnPtr							pending_mesh_connection;	///< used temporarily while establishing a mesh connection
		std::chrono::system_clock::time_point	last_updated;				///< when the connection was last used (TBD)
		std::chrono::system_clock::time_point	next_connection_attempt;	///< next time we can try to connect (for mesh networks)
		std::string								hostname;					///< the hostname as reported by the peer (may not be unique!)
		std::string								os_name;					///< the OS name/version as returned by EPS::get_os_name()
		std::string								eps_version;				///< the EPS version string for this client
		std::string								msg;						///< text (debug) message associated with this entry

		/** @todo I dislike the mix of std::chrono and time_t, but I'm not yet familiar enough with std::chrono to make
		 * it do exactly what I want.  At some point, this code will need to be cleaned up and we'll have to pick one
		 * standard for the entire codebase.  It is not yet that day.
		 */

		time_t									eps_uptime;					///< timestamp: how long EPS has been running
		time_t									system_uptime;				///< timestamp: how long this client has been running
		time_t									disconnect_timestamp;		///< what time this device disconnected

		/// Constructor.
		TrackerData(void) :
			active				(false						),
			mode				(EPS::kInvalid				),
			uuid				(boost::uuids::nil_uuid()	),
			connection			(nullptr					),
			eps_uptime			(0							),
			system_uptime		(0							),
			disconnect_timestamp(0							)
		{
			// ipv4, time_point and std::string all have a default constructor
			return;
		}

		/// Format the tracker data structure as a single text string.
		std::string as_debug(void) const;

		/// Determine if an entry is "active" or "inactive".  To be active, an entry must be available to send/receive packets. @{
		bool is_active(void) const
		{
			return	this->connection.get()				!= nullptr	&&
					this->connection->socket.is_open()	== true		&&
					this->active						== true		&&
					this->address_eps.is_unspecified()	== false	&&
					this->other_addresses.empty()		== false;
		}
		bool is_inactive(void) const { return not is_active(); }
		/// @}

		/// Determine if an entry is a client or server entry. @{
		bool is_client		(void) const { return this->mode == EPS::kClient; 	}
		bool is_server		(void) const { return this->mode == EPS::kServer;	}
		bool is_not_client	(void) const { return not this->is_client();		}
		bool is_not_server	(void) const { return not this->is_server();		}
		/// @}
	};

	/// Map between client UUIDs and a tracker record.  This is the central "database" that makes up the Tracker functionality.
	typedef std::map<UUID, EPS::TrackerData> TrackerMap;

	/// Map between an IP address and the corresponding UUID.  This is a secondary index to quickly find an entry given an IP address.
	typedef std::map<IPv4, UUID> DestinationMap;

	/// Manage all tracker records.
	class Tracker final
	{
		public:

			/// Add an entry but don't activate it.  For example, when the cache file is read from disk.
			EPS::TrackerData &add_inactive_entry(const UUID &uuid, const IPv4 &address_eps);

			/// Add (or update) a new tracker entry.
			EPS::TrackerData &add_active_entry(const UUID &uuid, const IPv4 &address_eps, const EPS::TConnPtr &connection);

			/// Activate an entry by setting the TCP connection pointer.
			EPS::TrackerData &activate_entry(const EPS::TConnPtr &connection);

			/// De-activate an entry by closing the TCP connection pointer and removing references to it in @ref eps_address_map  @{
			EPS::TrackerData &deactivate_entry(const EPS::TConnPtr &connection);
			EPS::TrackerData &deactivate_entry(const UUID &uuid);
			/// @}

			/// Update the timestamp for an entry so it doesn't get deleted.
			EPS::TrackerData &update_entry(const UUID &uuid);

			/// Quickly determine if a tracker entry exists for the given information. @{
			bool exists			(const UUID				&uuid		) const;
			bool exists			(const IPv4				&ipv4		) const;
			bool exists			(const TConnPtr			&connection	) const;
			bool does_not_exist	(const UUID				&uuid		) const { return not exists(uuid		); }
			bool does_not_exist	(const IPv4				&ipv4		) const { return not exists(ipv4		); }
			bool does_not_exist	(const TConnPtr			&connection	) const { return not exists(connection	); }
			/// @}

			/// Get a tracker record (create one if it doesn't already exist). @{
			EPS::TrackerData &get(const UUID &uuid);
			EPS::TrackerData &get(const IPv4 &eps_ipv4);
			EPS::TrackerData &get(const TConnPtr &connection);
			/// @}

			/** Create a new shared pointer for the given connection.  If a shared pointer already exists, then this
			 * will return the existing shared pointer, similar to @ref get_shared_pointer().
			 */
			EPS::TConnPtr create_shared_pointer(EPS::TCPConnection *connection);

			/** Look through all the shared pointers that Tracker knows about and see if we can find one matching the
			 * given raw pointer.  If one doesn't exist, then throw an exception. @see create_shared_pointer()
			 */
			EPS::TConnPtr get_shared_pointer(EPS::TCPConnection *connection);

			/// Disable entries that haven't been recently updated.
			void deactivate_old_entries(void);

			/// Determine if the tracker cache is empty.
			bool empty(void) const { return this->trk_map.empty(); }

			/// Get all data entries in the tracker map.
			const EPS::TrackerMap &get_all_entries(void) const { return this->trk_map; }

			/// Get the UUID which corresponds to the given destination IPv4 address.
			EPS::UUID uuid_for_destination(const IPv4 &ipv4) const;

			/// Return the set of enabled/valid connections. @{
			EPS::TCPConns get_all_connections(void) const;
			EPS::TCPConns get_all_connections_other_than(const TConnPtr &connection) const;
			/// @}

			/// Return the set of all UUIDs.
			EPS::UUIDSet get_all_uuids(void) const;

			/// Return the set of active UUIDs.
			EPS::UUIDSet get_all_active_uuids(void) const;

			/// Export the tracker information, so we remember which IP address was assigned to which client computer. @see EPS::Tracker::import_from_cache()
			void export_to_cache(const IPv4 &network) const;

			/// Import the tracker information (if any). @see EPS::Tracker::export_to_cache()
			void import_from_cache(const IPv4 &network);

		private:

			// members

			/// Map of all known connections, both active and inactive.  Maps UUIDs to TrackerData entries.
			EPS::TrackerMap trk_map;

			/// Map of all known active EPS IPv4 addresses.  Maps EPS IPs to UUIDs.  Does @b not map non-EPS addresses.
			EPS::DestinationMap eps_address_map;
	};
}
